package util.sql;

import java.util.Map;
import java.util.HashMap;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import org.apache.log4j.Category;

/**
 * Class that builds prepared statements using only a connection and an array of objects.
 */
public class PreparedStatementFactory {

    /**
     * The logging category for this class.
     */
    private static final Category cat = Category.getInstance(PreparedStatementFactory.class);

    /**
     * Given a prepared statment, set the given values in their corresponding statement indices.
     * 
     * @param stmnt  the statement to build
     * @param values the values to put into the statement
     * @exception IllegalArgumentException
     *                   when a null values array is passed in
     * @exception IllegalStateException
     *                   thrown when any SQLException occurs during the populating of the statement
     */
    public static void buildStatement(PreparedStatement stmnt,
                                      Object[] values) throws IllegalArgumentException, IllegalStateException {

        //If we have a null value here, then bail.
        if(values == null ||
           stmnt == null) {
            cat.error("The build statement function was passed a null value for the values array or the actual statement.  This is bad.");
            throw new IllegalArgumentException("Cannot pass a null value array into buildStatement().");
        }

        try {
            
            //Loop through each value, determine it's corresponding SQL type,
            //and stuff that value into the prepared statement.
            Object value = null;
            for(int i = 0; i < values.length; i++) {
                value = values[i];
    
                //Have to handle null values seperately
                if(value != null) {
                    
                    cat.debug("Setting the " + value + " value in the prepared statement at position " + (i + 1));

                    //If the object is our representation of a null value, then handle it seperately
                    if(value instanceof NullSQLType) {
                        stmnt.setNull(i + 1, ((NullSQLType) value).getFieldType());
                    } else {
                        stmnt.setObject(i + 1, value);
                    }
                } else {
                    //Can't do anything with a null value.
                    String message = "Can't use a null value in a prepared statement.  If you want to persist a null value, use the " +
                        "com.healthdec.data.NullSQLType object to represent this in the prepared statement.";
                    cat.error(message);
                    throw new IllegalStateException(message);
                }
            }
        } catch(SQLException e) {
            String message = "The prepared statement could not be formed with the given values";
            cat.error(message, e);
            throw new IllegalStateException(message);
        }
    }
}
