package util.sql;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.PreparedStatement;
import java.sql.Statement;
import java.sql.SQLException;
import org.apache.log4j.Category;

/**
 * Main class for performing all SQL operations, querying, updating, transactions etc...  The
 * main goal of this class is to take away the redundancy and overhead that occurs with each
 * SQL query/update.  The only variability in each SQL operation is:
 * <ul><li>the sql string
 * <li>the prepared statement values
 * <li>how the result set is handled</ul>
 * <p>
 * All other functionality, including getting a connection, entering values into a prepared
 * statement, exception handling etc... should factored out.  That's just what this class
 * does.
 * 
 * @author Ryan Daigle
 * @version 1.0
 * @since 12/3/2001
 */
public class SQLProcessor {

    /**
     * The logging category for this class.
     */
	private static Category cat = Category.getInstance(SQLProcessor.class);

    /**
     * Default constructor.  Private so no oustide instantiation.
     */
    private SQLProcessor() {
        //private so no outside instantiation
    }

	/**
	 * Get an instance of a SQLProcessor.  At this point it could be a singleton, or pooled etc...
	 * But you don't care, so always use this method to get an instance.
	 * 
	 * @return a SQLProcessor
	 */
	public static SQLProcessor getInstance() {
		return new SQLProcessor();
	}

	/**
	 * Start the execution of SQL query.
	 * 
	 * @param sql       the sql query string to execute
	 * @param pStmntValues
	 *                  the values to enter into the prepared statement.  If there are no values, use
	 *                  <code>null</code>
	 * @param processor the result processor that will handle the result set produced by this query.  This is what
	 *                  will ultimately decide what return value is handed back.
	 * @return the objects formed by the given result processor
	 */
	public Object[] executeQuery(String sql,
								  Object[] pStmntValues,
								  ResultProcessor processor) {

        //Get a connection (assume it's part of a ConnectionManager class)
        Connection conn = ConnectionManager.getConnection();

        //Hand of our connection to the method that will actually execute
		//the call
        Object[] results = handleQuery(sql, pStmntValues, processor, conn);

        //Close the connection
        closeConn(conn);

        //And return its results
        return results;
	}

	/**
	 * Method that handles the bulk of the query processing.  This method assumes that the given
	 * sql string is a valid sql statement.  This string will be formed into a prepared
	 * statement hand handled appropriately.  Calling members should always use the
	 * executeQuery method when complete statement and connection handling is desired!
	 * <p>
	 * Note: This method <i>does not</i> close the connection.  This is left up to the
	 * calling method as it is what supplied the connection. If all handling of the
	 * connections is desired, use the executeQuery method!  (However, in the case of
	 * any exceptions, the connection will closed)
	 * 
	 * @param sql       the sql query string to execute
	 * @param pStmntValues
	 *                  the values to enter into the prepared statement.  If there are no values, use
	 *                  <code>null</code>
	 * @param processor the result processor that will handle the result set produced by this query.  This is what
	 *                  will ultimately decide what return value is handed back.
	 * @param conn      the connection to use when getting a prepared statement
	 * @return the objects formed by the given result processor
	 * @exception DatabaseQueryException
	 */
    protected Object[] handleQuery(String sql,
                                   Object[] pStmntValues,
                                   ResultProcessor processor,
                                   Connection conn) {

		//Get a prepared statement to use
        PreparedStatement stmnt = null;

        try {

            //Get an actual prepared statement
            stmnt = conn.prepareStatement(sql);

            //Attempt to stuff this statement with the given values.  If
            //no values were given, then we can skip this step.
			cat.debug("Attempting to populate the prepared statement: " + sql);
            if(pStmntValues != null) {
                PreparedStatementFactory.buildStatement(stmnt, pStmntValues);
            }

            //Attempt to exceute the statement
			cat.debug("Attempting to execute the sql: " + sql);
			ResultSet rs = stmnt.executeQuery();
			cat.debug("Statement successfully executed");

            //Get the results from this query
            Object[] results = processor.process(rs);

            //Close out the statement only.  The conn will be closed by the
            //caller
            closeStmnt(stmnt);

            //Return the results.
			return results;

        //Any SQL Exceptions that occur should be recast to our runtime query
        //exception and thrown from here.
		} catch(SQLException e) {
			String message = "Could not perform the query for " + sql;
			cat.error(message, e);

            //Close out all resources on an exception
            closeConn(conn);
            closeStmnt(stmnt);

            //And rethrow as our runtime exception
			throw new DatabaseQueryException(message);
		}
    }

    /**
     * Execute an update and setup and destroy the connection associate with it.
     * 
     * @param sql       the sql query string to execute
     * @param pStmntValues
     *                  the values to enter into the prepared statement.  If there are no values, use <code>null</code>
     * @param processor the result processor that will handle the result set produced by this query.  This is what
     *                  will ultimately decide what return value is handed back.
     * @param appId     the app id of the application that's calling this method.  Used to get an app specific
     *                  connection.
     */
	public void executeUpdate(String sql,
							   Object[] pStmntValues,
							   UpdateProcessor processor) {

        //Get a connection
        Connection conn = ConnectionManager.getConnection();

        //Send it off to be executed
        handleUpdate(sql, pStmntValues, processor, conn);

        //Close the connection
        closeConn(conn);
	}

	/**
	 * Method that handles the bulk of the update processing.  This method assumes that the given
	 * sql string is a valid sql statement.  This string will be formed into a prepared
	 * statement hand handled appropriately.
	 * <p>
	 * Note: This method <i>does not</i> close the connection by default.  This is left up to the
	 * calling method as it is what supplied the connection.  It will only be closed in the case
	 * of an exception.
	 * 
	 * @param sql       the sql update string to execute
	 * @param pStmntValues
	 *                  the values to enter into the prepared statement.  If there are no values, use <code>null</code>
	 * @param processor
	 * @param conn
	 */
    protected void handleUpdate(String sql,
                                Object[] pStmntValues,
                                UpdateProcessor processor,
                                Connection conn) {

		//Get a prepared statement to use
        PreparedStatement stmnt = null;

        //Try to build the prepared statment with the given values.
		try {

            //Get an actual prepared statement
            stmnt = conn.prepareStatement(sql);

            //Attempt to stuff this statement with the given values.  If
            //no values were given, then we can skip this step.
			cat.debug("Attempting to build the statement for: " + sql);
            if(pStmntValues != null) {
                PreparedStatementFactory.buildStatement(stmnt, pStmntValues);
            }

            //Attempt to exceute the statement
			cat.debug("Attempting to execute the sql: " + sql);
			int rows = stmnt.executeUpdate();
            cat.debug("Statement successfully executed");

            //Now hand off the number of rows updated to the processor
			processor.process(rows);
            
            //Close out the statement only.  The conn will be closed by the
            //caller
            closeStmnt(stmnt);

        //Any SQL Exceptions that occur should be recast to our runtime query
        //exception and thrown from here.
		} catch(SQLException e) {
			String message = "Could not perform the update for " + sql;
			cat.error(message, e);

            //Close out all resources on an exception
            closeConn(conn);
            closeStmnt(stmnt);

            //And rethrow as our exception
			throw new DatabaseQueryException(message);
		}
    }

	/**
	 * Close the connection, can handle null connections.
	 * 
	 * @param conn   the connection to close
	 */
	public void closeConn(Connection conn) {
		if(conn != null) {
			try {
				conn.close();
			} catch(SQLException e) {
				//swallow exception.  Since we're closing it we'll let it be dead.
			}
		}
	}

	/**
	 * Close the statement, can handle null statements.
	 * 
	 * @param conn   the statement to close
	 */
	private void closeStmnt(Statement stmnt) {
		if(stmnt != null) {
			try {
				stmnt.close();
			} catch(SQLException e) {
				//swallow exception.  Since we're closing it we'll let it be dead.
			}
		}
	}

    /**
     * Start a transaction.
     * 
     * @param appId  the app id to use when determining the connection's permissions
     * @return the transaction object to use when performing queries and updates that are a part of this
     *         transaction.
     */
    public SQLTransaction startTransaction() {
        return new SQLTransaction(this);
    }
}