package util.sql;

import java.sql.SQLException;
import java.sql.Connection;
import org.apache.log4j.Category;

/**
 * A SQL transaction.  Uses a SQL Processor to execute queries and updates as part of the
 * same transaction.
 * 
 * @author Ryan Daigle
 * @version 1.0
 * @since 12/03/2001
 */
public class SQLTransaction {

    /**
     * The logging category for this class.
     */
    private static Category cat = Category.getInstance(SQLTransaction.class);

    /**
     * The SQL Processor to use for this transaction.
     */
    private SQLProcessor sqlProcessor;

    /**
     * The connection to use for this transaction.
     */
    private Connection conn;

    /**
     * Default constructor.  Initializes a new connection for this transaction only.
     * 
     * @param appId     the app id that will determine this transaction's connection permissions
     * @param processor
     */
    public SQLTransaction(SQLProcessor sqlProcessor) {

        //Set the processor and get a connection
        this.sqlProcessor = sqlProcessor;

        //And start a transaction
        startTransaction();
    }

	/**
	 * Start a transaction.  This is automatically called on instantiation, so it only needs to be
	 * called after a transaction has been committed and a new one needs to start.  It
	 * initializes the member connection only if it's <code>null</CODE>.
	 */
    public void startTransaction() {

        //Get a new connection if the current one is null.
		if(conn == null) {
			conn = ConnectionManager.getConnection();
		}

        //Try to turn auto commit off.  If fails, try and proceed anyways.
        try {
            cat.debug("Attempting to turn off auto commit on this connection.");
            conn.setAutoCommit(false);
        } catch(SQLException e) {
            cat.error("Could not turn off auto commit on this connection.  Attempting to procee with transaction, but cannot guarantee transactional integrity.", e);
        }
    }

	/**
	 * Execute a query as part of this transaction.  If any exception occurs as part
	 * of this execution, then rollback all queries/updates up to this point
	 * and rethrow the Database exception.
	 * 
	 * @param sql       the sql statement to execute
	 * @param pStmntValues
	 *                  the values to fill the prepared statement with, can be <code>null</code> if there are no
	 *                  values to fill
	 * @param processor the processor that will handle the result set
	 * @return the objects returned from the result processor
	 */
    public Object[] executeQuery(String sql,
                                 Object[] pStmntValues,
                                 ResultProcessor processor) {

        //Try and get the results.  If a query fails, then rollback
        //the transaction and rethrow the exception.
        try {
            return sqlProcessor.handleQuery(sql,
                                            pStmntValues,
                                            processor,
                                            conn);
        } catch(DatabaseQueryException e) {
            cat.error("Error as part of the transaction.  Attempting to rollback the transaction.", e);
            rollbackTransaction();
            throw e;
        } 
    }

	/**
	 * Execute an update as part of this transaction. If any exception occurs as part
	 * of this execution, then rollback all queries/updates up to this point and
	 * rethrow the Database exception.
	 * 
	 * @param sql       the sql statement to execute
	 * @param pStmntValues
	 *                  the values to fill the prepared statement with, can be <code>null</code> if there are no
	 *                  values to fill
	 * @param processor
	 */
    public void executeUpdate(String sql,
                              Object[] pStmntValues,
                              UpdateProcessor processor) {

        //Try and get the results.  If a update fails, then rollback
        //the transaction and rethrow the exception.
        try {
            sqlProcessor.handleUpdate(sql,
                                      pStmntValues,
                                      processor,
                                      conn);

        } catch(DatabaseUpdateException e) {
            cat.error("Error as part of the transaction.  Attempting to rollback the transaction.", e);
            rollbackTransaction();
            throw e;
        } 
    }

    /**
     * Commit all queries and updates that are a part of this transaction.  If any thing fails in
     * this step, then rollback the transaction.
     */
    public void commitTransaction() {

        //Try to commit and release all resources
        try {
            cat.debug("Attempting to commit the current transaction.");
            conn.commit();
            sqlProcessor.closeConn(conn);
            cat.debug("Successfully committed current transaction.");

        //If something happens then attempt a rollback and release resources.
        } catch(Exception e) {
            cat.error("Could not commit this transaction, attempting to rollback the transaction.", e);
            rollbackTransaction();
            throw new DatabaseUpdateException("Could not commit the current transaction.");
        }
    }

    /**
     * Rollback all queries and updates that are a part of this transaction.  If any thing fails in
     * this step, then just close the connection and swallow the exception.
     */
    private void rollbackTransaction() {

        //Try to commit and release all resources
        try {
            cat.debug("Attempting to rollback the current transaction.");
            conn.rollback();
            conn.setAutoCommit(true);
            sqlProcessor.closeConn(conn);
            cat.debug("Successfully rolled the current transaction back.");

        //If something happens then just swallow it
        } catch(SQLException e) {
            cat.error("Could not rollback this transaction.", e);
            sqlProcessor.closeConn(conn);
        }
    }
}
