package util.sql.test;

import util.sql.SQLProcessor;
import util.sql.SQLTransaction;
import util.sql.ResultProcessor;
import util.sql.UpdateProcessor;
import util.sql.MandatoryUpdateProcessor;

/**
 * Test driver to test all functionality of the framework.<p>
 * <I>Usage:</I> java util.sql.test.TestClient -cp ./sqlFramework.jar;./log4j.jar;./yourDbDriver.jar get |
 * getAll | update | updateAll | failTransaction
 * <P>
 * <I>Assumes:</I> a USERS table with columns:
 * <li>ID
 * <li>USERNAME
 * <li>F_NAME
 * <li>L_NAME
 * <li>EMAIL
 * 
 * @author Ryan Daigle
 * @version 1.0
 * @since 3/19/2002
 */
public class TestClient {

	//The SQL statements
	private static final String SQL_GET_USER = "SELECT * FROM USERS WHERE ID = ?";
	private static final String SQL_GET_ALL_USERS = "SELECT * FROM USERS";
	private static final String SQL_UPDATE_USER = "UPDATE USERS SET USERNAME = ?, F_NAME = ?, L_NAME = ?, EMAIL = ? WHERE ID = ?";

	//The sql and result processors
	private static SQLProcessor sqlProcessor = SQLProcessor.getInstance();
	private static ResultProcessor userProcessor = UserResultProcessor.getInstance();
	private static UpdateProcessor updateProcessor = MandatoryUpdateProcessor.getInstance();

	//The test commands
	private static final String COMMAND_GET = "get";
	private static final String COMMAND_GET_ALL = "getAll";
	private static final String COMMAND_UPDATE = "update";
	private static final String COMMAND_UPDATE_ALL = "updateAll";
	private static final String COMMAND_FAIL_TRANSACTION = "failTransaction";
	private static final String COMMAND_NONE = "none";

	/**
	 * The test driver for the sql processor.
	 * 
	 * @param args   the test command
	 */
	public static void main(String[] args) {

		//Get the test command
        String command = (args.length == 0 ? COMMAND_NONE : args[0]);

		//Catch all database runtime exceptions
		try {

			initializeTestDB();

			//Perform the proper test case based on the command
			if(COMMAND_GET.equals(command)) {
				User user = getUser(1);
				System.out.println("Retrieved user: " + user);
	
			} else if(COMMAND_GET_ALL.equals(command)) {
				printAll();
	
			} else if(COMMAND_UPDATE.equals(command)) {
				updateUser(new User(1, "updatedUser", "Updated", "User", "updated@test.com"));
				printAll();
	
			} else if(COMMAND_UPDATE_ALL.equals(command)) {
				updateUsers(new User[] {new User(1, "updatedUser1", "Updated1", "User1", "updated1@test.com"),
					new User(2, "updatedUser2", "Updated2", "User2", "updated2@test.com")});
				printAll();
	
			} else if(COMMAND_FAIL_TRANSACTION.equals(command)) {
				
				//Print out the state before the attempted transaction
                printAll();

				try {
					updateUsers(new User[] {new User(1, "updatedUser3", "Updated3", "User3", "updated3@test.com"),
						new User(9871, "updatedUser9871", "Updated9871", "User9871", "updated9871@test.com"),
						new User(2, "updatedUser4", "Updated4", "User4", "updated4@test.com")});
				} catch(Exception e) {
					System.out.println("Transaction threw an exception.");
					e.printStackTrace(System.out);
				}

                printAll();
	
			} else if(COMMAND_NONE.equals(command)) {
				System.out.println("Usage: \"util.sql.test.TestClient get | getAll | update | updateAll | failTransaction\"");
			}
		} catch(Exception e) {
			e.printStackTrace(System.out);
		}
	}

	/**
	 * Print the users table to the standard out.
	 */
	private static void printAll() {
		System.out.println("Current DB state:");
		User[] users = getAllUsers();
		User user = null;
		for (int i = 0; i < users.length; i++) {
			user = users[i];
			System.out.println("User: " + user);
		}
	}

	/**
	 * Print the users table to the standard out.
	 */
	private static void initializeTestDB() {
		System.out.println("Resetting test db values.");
		updateUsers(new User[] {new User(1, "rdaigle", "Ryan", "Daigle", "rdaigle@test.com"),
			new User(2, "sroberts", "Scott", "Roberts", "sroberts@test.com"),
			new User(3, "rfarris", "Rick", "Farris", "rfarris@test.com"),
			new User(4, "lreynolds", "Lori", "Reynolds", "lreynolds@test.com")});
	}

	/**
	 * Get a user with the given user id.
	 * 
	 * @param userId the id of the user to get
	 * @return the user with the given id
	 */
	public static User getUser(int userId) {

		//Use our get user by id SQL statement plugging in the given userId as the sole
		//argument of the resulting prepared statement.  Use the generic user result
		//processor to collect the results.
		Object[] users = sqlProcessor.executeQuery(SQL_GET_USER,
												   new Object[] {new Integer(userId)},
												   userProcessor);

		//Since we know only one user was returned as part of this query, just
		//return the first one (Even if it's null)
		return (User) users[0];
	}

	/**
	 * Get all users
	 * 
	 * @return all users
	 */
	public static User[] getAllUsers() {

		//Use our get all users SQL statement.  Reuse the generic user result
		//processor to collect the results.
		Object[] users = sqlProcessor.executeQuery(SQL_GET_ALL_USERS,
												   null,
												   userProcessor);

		//Return all users
		return (User[]) users;
	}

	/**
	 * Get all users
	 * 
	 * @return all users
	 */
	public static void updateUser(User user) {

		//Use our get all users SQL statement.  Reuse the generic user result
		//processor to collect the results.
		sqlProcessor.executeUpdate(SQL_UPDATE_USER,
								   new Object[] {user.getUserName(),
									   user.getFirstName(),
									   user.getLastName(),
									   user.getEmail(),
									   new Integer(user.getId())},
								   updateProcessor);
	}

	/**
	 * Update all these users.  Merely an example of a transaction.
	 * 
	 * @param users  the array of users to update
	 */
	public static void updateUsers(User[] users) {

		//Get our transaction
		SQLTransaction trans = sqlProcessor.startTransaction();
		
		//For each user, update it
		User user = null;
		for(int i = 0; i < users.length; i++) {
			user = users[i];
			trans.executeUpdate(SQL_UPDATE_USER,
								new Object[] {user.getUserName(),
									user.getFirstName(),
									user.getLastName(),
									user.getEmail(),
									new Integer(user.getId())},
								updateProcessor);
		}

		//Now commit the transaction
		trans.commitTransaction();
	}
}
